#include "freelist_internal.h"





/****************************************************************************/
void freelist_query( struct freelist_state *fs, enum freelist_query_type query_type, void *query_input, void *query_output )
{
  assert( fs != NULL );
  // TRD : query type can be any value in its range
  // TRD : query_input can be NULL in some cases
  assert( query_output != NULL );

  switch( query_type )
  {
    case FREELIST_QUERY_ELEMENT_COUNT:
      assert( query_input == NULL );

      *(atom_t *) query_output = fs->element_count;
    break;

    case FREELIST_QUERY_VALIDATE:
      assert( query_input != NULL );

      freelist_internal_validate( fs, (atom_t) query_input, (enum freelist_validity *) query_output );
    break;
  }

  return;
}





/****************************************************************************/
void freelist_internal_validate( struct freelist_state *fs, atom_t expected_number_elements, enum freelist_validity *fv )
{
  struct freelist_element
    *fe;

  atom_t
    element_count = 0;

  assert( fs != NULL );
  // TRD : expected_number_elements can be any value in its range
  assert( fv != NULL );

  *fv = FREELIST_VALID;

  fe = (struct freelist_element *) fs->top[FREELIST_POINTER];

  while( fe != NULL and element_count < expected_number_elements )
  {
    element_count++;
    fe = (struct freelist_element *) fe->next[FREELIST_POINTER];
  }

  /* TRD : once we reach the expected number of elements, the next
           pointer of the final element should be NULL

           if it is not, we have either extra elements or a loop

           we may however be short elements, in which case fe will
           be NULL before element_count reaches expected_number_elements

           if we have more elements than we expected,
           we either actually have excess elements
           or we have a loop

           to detect the loop, we leave fe pointing where it is
           and run a second pointer down the freelist for element_count
           number of steps

           if we 'see' fe prior to reaching element_count steps
           we must have had a loop

           note that if an element is present twice in a freelist,
           we must have a loop
  */

  if( fe != NULL and fe->next[FREELIST_POINTER] != NULL )
  {
    struct freelist_element
      *temp_fe;

    atom_t
      temp_count = 0;

    *fv = FREELIST_INVALID_ADDITIONAL_ELEMENTS;

    temp_fe = (struct freelist_element *) fs->top[FREELIST_POINTER];

    while( temp_count < element_count and temp_fe != fe )
    {
      temp_count++;
      temp_fe = (struct freelist_element *) temp_fe->next[FREELIST_POINTER];
    }

    if( temp_fe == fe and temp_count < element_count )
      *fv = FREELIST_INVALID_LOOP;
  }

  if( element_count < expected_number_elements )
    *fv = FREELIST_INVALID_MISSING_ELEMENTS;

  return;
}

