#include "queue_internal.h"





/****************************************************************************/
#pragma warning( disable : 4100 )

void queue_query( struct queue_state *qs, enum queue_query_type query_type, void *query_input, void *query_output )
{
  assert( qs != NULL );
  // TRD : query_type can be any value in its range
  // TRD : query_input can be NULL
  assert( query_output != NULL );

  switch( query_type )
  {
    case QUEUE_QUERY_ELEMENT_COUNT:
      assert( query_input == NULL );

      freelist_query( qs->fs, FREELIST_QUERY_ELEMENT_COUNT, NULL, query_output );
    break;

    case QUEUE_QUERY_VALIDATE:
      assert( query_input != NULL );

      queue_internal_validate( qs, (atom_t) query_input, (enum queue_validity *) query_output );
    break;
  }

  return;
}

#pragma warning( default : 4100 )





/****************************************************************************/
void queue_internal_validate( struct queue_state *qs, atom_t expected_number_elements, enum queue_validity *qv )
{
  struct queue_element
    *qe;

  atom_t
    element_count = 0;

  assert( qs != NULL );
  // TRD : expected_number_elements can be any value in its range
  assert( qv != NULL );

  /* TRD : remember there is a dummy element in the queue */
  expected_number_elements++;

  *qv = QUEUE_VALID;

  qe = (struct queue_element *) qs->dequeue[QUEUE_POINTER];

  while( qe != NULL and element_count < expected_number_elements )
  {
    element_count++;
    qe = (struct queue_element *) qe->next[QUEUE_POINTER];
  }

  /* TRD : once we reach the expected number of elements, the next
           pointer of the final element should be NULL

           if it is not, we have either extra elements or a loop

           we may however be short elements, in which case qe will
           be NULL before element_count reaches expected_number_elements

           if we have more elements than we expected,
           we either actually have excess elements
           or we have a loop

           to detect the loop, we leave qe pointing where it is
           and run a second pointer down the queue for element_count
           number of steps

           if we 'see' qe prior to reaching element_count steps
           we must have had a loop

           note that if an element is present twice in a queue,
           we must have a loop
  */

  if( qe != NULL and qe->next[QUEUE_POINTER] != NULL )
  {
    struct queue_element
      *temp_qe;

    atom_t
      temp_count = 0;

    *qv = QUEUE_INVALID_ADDITIONAL_ELEMENTS;

    temp_qe = (struct queue_element *) qs->dequeue[QUEUE_POINTER];

    while( temp_count < element_count and temp_qe != qe )
    {
      temp_count++;
      temp_qe = (struct queue_element *) temp_qe->next[QUEUE_POINTER];
    }

    if( temp_qe == qe and temp_count < element_count )
      *qv = QUEUE_INVALID_LOOP;
  }

  if( element_count < expected_number_elements )
    *qv = QUEUE_INVALID_MISSING_ELEMENTS;

  return;
}

