#include "ringbuffer_internal.h"





/****************************************************************************/
void ringbuffer_delete( struct ringbuffer_state *rs, void (*user_data_delete_function)(void *user_data, void *user_state), void *user_state )
{
  struct ringbuffer_element
    *re;

  assert( rs != NULL );
  // TRD : user_data_delete_function can be NULL
  // TRD : user_state can be NULL

  /* TRD : fully dequeue
           this returns all elements to the freelist
           (e.g. by fully dequeuing the ringbuffer)
           then delete the ringbuffer and freelist
  */

  while( ringbuffer_internal_dequeue_element(rs, &re) )
    if( user_data_delete_function != NULL )
      user_data_delete_function( re->user_data, user_state );

  /* TRD : fully dequeuing will leave us
           with a single dummy element in the queue
           which both qs->enqueue and qs->dequeue point at
           we push this back onto the freelist
           before we delete the freelist
  */

  freelist_push( rs->fs, rs->enqueue[RINGBUFFER_POINTER]->fe );

  freelist_delete( rs->fs, ringbuffer_internal_freelist_delete_function, NULL );

  abstraction_aligned_free( rs );

  return;
}





/****************************************************************************/
#pragma warning( disable : 4100 )

void ringbuffer_internal_freelist_delete_function( void *user_data, void *user_state )
{
  assert( user_data != NULL );
  assert( user_state == NULL );

  abstraction_aligned_free( user_data );

  return;
}

#pragma warning( default : 4100 )

