/***** ANSI includes *****/
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/***** internal includes *****/
#include "abstraction.h"

/***** external includes *****/
#include "liblfds.h"

/***** defines *****/
#define and &&
#define or  ||

#define RAISED   1
#define LOWERED  0

#define NO_FLAGS 0x0

/***** enums *****/
enum test_operation
{
  UNKNOWN,
  HELP,
  TEST,
  BENCHMARK
};

enum error
{
  ERROR_NO_ERROR,
  ERROR_ADDITIONAL_ELEMENTS,
  ERROR_MISSING_ELEMENTS,
  ERROR_LOOP,
  ERROR_TEST_DATA
};

/***** structs *****/
#include "structures.h"

/***** prototypes *****/
int main( int argc, char **argv );

void internal_display_test_name( char *test_name );
void internal_display_test_result( enum error e );

void benchmark_queue( void );
  thread_return_t CALLING_CONVENTION queue_internal_thread_benchmark( void *queue_benchmark );

void test_abstraction( void );
  void abstraction_test_increment( void );
    thread_return_t CALLING_CONVENTION abstraction_test_internal_thread_increment( void *shared_counter );
    thread_return_t CALLING_CONVENTION abstraction_test_internal_thread_atomic_increment( void *shared_counter );
  void abstraction_test_dcas( void );
    thread_return_t CALLING_CONVENTION abstraction_test_internal_thread_dcas( void *abstraction_test_dcas_state );

void test_freelist( void );
  void freelist_test_internal_popping( void );
    int freelist_test_internal_popping_init( void **user_data, void *user_state );
    thread_return_t CALLING_CONVENTION freelist_test_internal_thread_popping( void *freelist_test_popping_state );
  void freelist_test_internal_pushing( void );
    int freelist_test_internal_pushing_init( void **user_data, void *user_state );
    void freelist_test_internal_pushing_delete( void *user_data, void *user_state );
    thread_return_t CALLING_CONVENTION freelist_test_internal_thread_pushing( void *freelist_test_pushing_state );
  void freelist_test_internal_popping_and_pushing( void );
    thread_return_t CALLING_CONVENTION freelist_test_internal_thread_popping_and_pushing_start_popping( void *freelist_test_popping_and_pushing_state );
    thread_return_t CALLING_CONVENTION freelist_test_internal_thread_popping_and_pushing_start_pushing( void *freelist_test_popping_and_pushing_state );
  void freelist_test_internal_rapid_popping_and_pushing( void );
    thread_return_t CALLING_CONVENTION freelist_test_internal_thread_rapid_popping_and_pushing( void *freelist_state );
  void freelist_test_internal_convert_freelist_validity_to_error( enum freelist_validity fv, enum error *e );

void test_queue( void );
  void queue_test_enqueuing( void );
    thread_return_t CALLING_CONVENTION queue_test_internal_thread_simple_enqueuer( void *queue_test_enqueuing_state );
  void queue_test_dequeuing( void );
    thread_return_t CALLING_CONVENTION queue_test_internal_thread_simple_dequeuer( void *queue_test_dequeuing_state );
  void queue_test_enqueuing_and_dequeuing( void );
    thread_return_t CALLING_CONVENTION queue_test_internal_thread_enqueuer( void *queue_test_enqueuing_and_dequeuing_state );
    thread_return_t CALLING_CONVENTION queue_test_internal_thread_dequeuer( void *queue_state );
  void queue_test_rapid_enqueuing_and_dequeuing( void );
    thread_return_t CALLING_CONVENTION queue_test_internal_thread_rapid_enqueuer_and_dequeuer( void *queue_state );
  void queue_test_internal_convert_queue_validity_to_error( enum queue_validity qv, enum error *e );

void test_ringbuffer( void );
  thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_reader( void *ringbuffer_state );
  thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_writer( void *ringbuffer_state );

void test_slist( void );
  thread_return_t CALLING_CONVENTION slist_internal_thread_head_writer( void *slist_thread_start_state );
  thread_return_t CALLING_CONVENTION slist_internal_thread_after_writer( void *slist_thread_start_state );
  thread_return_t CALLING_CONVENTION slist_internal_thread_traverser( void *slist_thread_start_state );
  thread_return_t CALLING_CONVENTION slist_internal_thread_deleter_traverser( void *slist_thread_start_state );

void test_stack( void );
  thread_return_t CALLING_CONVENTION stack_internal_thread_reader( void *stack_state );
  thread_return_t CALLING_CONVENTION stack_internal_thread_writer( void *stack_state );

