#include "internal.h"





/****************************************************************************/
void test_ringbuffer( void )
{
  unsigned int
    loop,
    cpu_count;

  struct ringbuffer_state
    *rs;

  atom_t
    ringbuffer_count;

  thread_state_t
    *thread_handles;

  /* TRD : there are five tests

           1. one reader thread per CPU
              - ringbuffer always empty
           2. one writer thread per CPU
              - ringbuffer always growing
           3. one reader and one writer thread per CPU
              - ringbuffer balanced
           4. two reader and one writer thread per CPU
              - ringbuffer tending to be empty
           5. one reader and two writers thread per CPU
              - ringbuffer tending to grow
  */

  // TRD : to get the number of CPUs
  cpu_count = abstraction_cpu_count();

  printf( "\n"
          "Ringbuffer Test\n"
          "===============\n" );

  // TRD : 1. one reader thread per CPU

  printf( "\n"
          "1. one reader thread per CPU\n"
          "============================\n" );

  ringbuffer_count = ringbuffer_new( &rs, 10000, NULL, NULL );

  if( rs == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = (thread_state_t *) malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], loop, ringbuffer_internal_thread_reader, rs );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( rs, NULL, NULL );

  free( thread_handles );

  // TRD : 2. one writer thread per CPU

  printf( "\n"
          "2. one writer thread per CPU\n"
          "============================\n" );

  ringbuffer_count = ringbuffer_new( &rs, 5, NULL, NULL );

  if( rs == NULL or ringbuffer_count != 5 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = (thread_state_t *) malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], loop, ringbuffer_internal_thread_writer, rs );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( rs, NULL, NULL );

  free( thread_handles );

  // TRD : 3. one reader and one writer thread per CPU

  printf( "\n"
          "3. one reader and one writer thread per CPU\n"
          "===========================================\n" );

  ringbuffer_count = ringbuffer_new( &rs, 10000, NULL, NULL );

  if( rs == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = (thread_state_t *) malloc( sizeof(thread_state_t) * cpu_count * 2 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], loop, ringbuffer_internal_thread_reader, rs );
    abstraction_thread_start( &thread_handles[loop+cpu_count], loop, ringbuffer_internal_thread_writer, rs );
  }

  for( loop = 0 ; loop < cpu_count * 2 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( rs, NULL, NULL );

  free( thread_handles );

  // TRD : 4. two reader and one writer thread per CPU

  printf( "\n"
          "4. two reader and one writer thread per CPU\n"
          "===========================================\n" );

  ringbuffer_count = ringbuffer_new( &rs, 10000, NULL, NULL );

  if( rs == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = (thread_state_t *) malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], loop, ringbuffer_internal_thread_reader, rs );
    abstraction_thread_start( &thread_handles[loop+cpu_count], loop, ringbuffer_internal_thread_reader, rs );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], loop, ringbuffer_internal_thread_writer, rs );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( rs, NULL, NULL );

  free( thread_handles );

  // TRD : 5. one reader and two writers thread per CPU

  printf( "\n"
          "5. one reader and two writers thread per CPU\n"
          "============================================\n" );

  ringbuffer_count = ringbuffer_new( &rs, 10000, NULL, NULL );

  if( rs == NULL or ringbuffer_count != 10000 )
  {
    printf( "ringbuffer_new returned %llu\n", (unsigned long long int) ringbuffer_count );
    return;
  }

  thread_handles = (thread_state_t *) malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], loop, ringbuffer_internal_thread_reader, rs );
    abstraction_thread_start( &thread_handles[loop+cpu_count], loop, ringbuffer_internal_thread_writer, rs );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], loop, ringbuffer_internal_thread_writer, rs );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  ringbuffer_delete( rs, NULL, NULL );

  free( thread_handles );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_reader( void *ringbuffer_state )
{
  struct ringbuffer_state
    *rs;

  struct ringbuffer_element
    *re;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( ringbuffer_state != NULL );

  rs = (struct ringbuffer_state *) ringbuffer_state;

  time( &start_time );

  while( time(NULL) < start_time + 10 )
    if( ringbuffer_get_current_read_element(rs, &re) )
    {
      ringbuffer_put_read_element( rs, re );
      count++;
    }

  printf( "read count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION ringbuffer_internal_thread_writer( void *ringbuffer_state )
{
  struct ringbuffer_state
    *rs;

  struct ringbuffer_element
    *re;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( ringbuffer_state != NULL );

  rs = (struct ringbuffer_state *) ringbuffer_state;

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    ringbuffer_get_current_write_element( rs, &re, NULL );
    ringbuffer_put_write_element( rs, re );

    count++;
  }

  printf( "write count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}

